#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

int serial_file=-1;

int serial_init(const char *serial_device)
{
   // open the serial device (not a controlling terminal)
   serial_file=open(serial_device, O_RDWR | O_NOCTTY);

   if (serial_file<0)
   {
      // open failed
      return 0;
   }

   // start with clean new settings
   struct termios new_serial_settings;
   memset(&new_serial_settings, 0, sizeof(new_serial_settings));

   // 9600 baud 8N1, read enabled, no modem control signals needed
   new_serial_settings.c_cflag=(B9600 | CS8 | CLOCAL | CREAD);

   // raw output
   new_serial_settings.c_oflag=0;

   // turn off canonical input
   new_serial_settings.c_lflag=0;

   // timeout of 400 ms
   new_serial_settings.c_cc[VMIN]  = 0x0;
   new_serial_settings.c_cc[VTIME] = 20;

   // flush the port
   tcflush(serial_file, TCIOFLUSH);

   // apply new settings to the serial port
   tcsetattr(serial_file, TCSANOW, &new_serial_settings);

   return 1;
}

void serial_reset()
{
   printf("(resetting serial port) ");
   if (serial_file>=0)
   {
      // flush the port
      tcflush(serial_file, TCIOFLUSH);

      sleep(4);

      // start with clean new settings
      struct termios new_serial_settings;
      memset(&new_serial_settings, 0, sizeof(new_serial_settings));

      // 9600 baud 8N1, read enabled, no modem control signals needed
      new_serial_settings.c_cflag=(B9600 | CS8 | CLOCAL | CREAD);

      // raw output
      new_serial_settings.c_oflag=0;

      // turn off canonical input
      new_serial_settings.c_lflag=0;

      // timeout of 400 ms
      new_serial_settings.c_cc[VMIN]  = 0x0;
      new_serial_settings.c_cc[VTIME] = 20;

      // flush the port
      tcflush(serial_file, TCIOFLUSH);

      // apply new settings to the serial port
      tcsetattr(serial_file, TCSANOW, &new_serial_settings);
   }
}

// verify 4 byte header of the reply, return zero if checksum fails
uint8_t check_header(const uint8_t* buffer, int x)
{
   uint8_t cs=0;
   while(x--)
   {
      cs^=*(buffer++);
   }
   return(!cs);
}

// verify x byte data section of the reply, return zero if checksum fails
uint8_t check_payload(const uint8_t* buffer, int x)
{
   uint8_t cs=0;
   if (x>4)
   {
      x-=4;
      buffer+=4;
   }

   while(x--)
   {
      cs^=*(buffer++);
   }
   return(!cs);
}


// wait for reply_size bytes, return number of bytes placed in rx_buffer
int read_reply(uint8_t *rx_buffer, int reply_size)
{
   int rx_count=0;
   while (rx_count<reply_size)
   {
      int x=read(serial_file, rx_buffer, reply_size-rx_count);
      if (0==x)
      {
         // timeout!
         return(0);
      }
      rx_count+=x;
      rx_buffer+=x;
   }
   return(rx_count);
}

// transmit a 4 byte command to the FE-5680A, wait for a reply, verify header and data reply
// return value is the number of bytes placed in the rx_buffer
int tx_command(uint8_t command, uint8_t *rx_buffer)
{
   uint8_t tx_buffer[4];
   uint8_t reply_length;

   tx_buffer[0]=command;
   tx_buffer[1]=4;
   tx_buffer[2]=0;
   tx_buffer[3]=tx_buffer[0] ^ tx_buffer[1] ^ tx_buffer[2];    // calc checksum
   write(serial_file, tx_buffer, sizeof(tx_buffer));

   if (0x04==read_reply(rx_buffer, 4))
   {
      // header received, check length and receive the data
      if (check_header(rx_buffer, 4))
      {
         // second byte is data payload length
         reply_length=rx_buffer[1]-4;
         rx_buffer+=4;

         if (reply_length==read_reply(rx_buffer, reply_length))
         {
            if (!check_payload(rx_buffer-4, reply_length+4))
            {
               // checksum was wrong, so note it (but don't reset serial port)
               printf("(CS bad!) ");
            }
            return(reply_length+4);
         }
      }
   }
   serial_reset();
   return(0);
}


void test_command(uint8_t command)
{
   int reply_length;
   int i;
   uint8_t rx_buffer[256];

   printf("\n\nCmd 0x%2.2X ", command);
   reply_length=tx_command(command, rx_buffer);
   printf("0x%2.2X byte reply: ", reply_length);

   // show header+data+checksum in the hex dump
   for (i=0; i<reply_length; ++i)
   {
      printf("[%2.2X] ", rx_buffer[i]);
   }

   // but for the ascii output, stop before the data checksum
   reply_length-=1;

   if (reply_length>4)
   {
      // and start printing ascii after the header
      printf("\nCmd 0x%2.2X ASCII (.): ", command);
      for (i=4; i<reply_length; ++i)
      {
         if ((rx_buffer[i]>=' ') && (rx_buffer[i]<='~'))
         {
            printf("%c", rx_buffer[i]);
         }
         else
         {
            // non printables replaced with a decimal point
            printf(".");
         }
      }

      printf("\nCmd 0x%2.2X ASCII ( ): ", command);
      for (i=4; i<reply_length; ++i)
      {
         if ((rx_buffer[i]>=' ') && (rx_buffer[i]<='~'))
         {
            printf("%c", rx_buffer[i]);
         }
         else
         {
            // non printables replaced with a space
            printf(" ");
         }
      }
   }
}


int main(int argc, const char *argv[])
{
   if (argc<2)
   {
      printf("\nUsage: %s comport\n", argv[0]);
      return(1);
   }

   if (0==serial_init(argv[1]))
   {
      printf("\nCouldn't open port %s\n", argv[1]);
      return(1);
   }

   // based on my two working units:
   // dynamic: appears to change while the unit is in operation (ADC readings, etc.)
   // static: doesn't appear to change while the unit is operating
   // unique: different replies from different units (like a unit serial number)
   // non-unique: same reply from different units

   test_command(0x22);     // dynamic
   test_command(0x29);     // non-unique, appears static
   test_command(0x2b);     // unique, appears to be static, variable length reply, might be asciiz text (">20000000.000000")
   test_command(0x2d);     // readback of 32 bit value of dds offset (documented!)
   test_command(0x47);     // non-unique, appears static
   test_command(0x53);     // dynamic, 8 bit value, related to temperature?
   test_command(0x57);     // unique, I've seen it change after modifying the DDS offset
   test_command(0x59);     // unique, I've seen it change after modifying the DDS offset
   test_command(0x5a);     // dynamic, 4 channels of 12 bit ADC data?
   test_command(0x61);     // unique, appears static, asciiz unit number from barcode?
   test_command(0x65);     // unique, reply data sometimes has a bad checksum
   test_command(0x67);     // unique
   test_command(0xf0);     // asciiz version number?
   return(0);
}
